@@ -2,7 +2,7 @@
 // It is subject to the license terms in the LICENSE file found in the top-level directory
 // of this distribution and at http://opencv.org/license.html.
 //
-// Copyright (C) 2018 Intel Corporation
+// Copyright (C) 2018-2019 Intel Corporation
 
 
 #ifndef OPENCV_GAPI_GKERNEL_HPP
@@ -14,7 +14,6 @@
 #include <type_traits> // false_type, true_type
 #include <unordered_map> // map (for GKernelPackage)
 #include <utility> // tuple
-#include <vector>  // lookup order
 
 #include <opencv2/gapi/gcommon.hpp> // CompileArgTag
 #include <opencv2/gapi/util/util.hpp> // Seq
@@ -23,7 +22,7 @@
 #include <opencv2/gapi/gmetaarg.hpp>  // GMetaArg
 #include <opencv2/gapi/gtype_traits.hpp> // GTypeTraits
 #include <opencv2/gapi/util/compiler_hints.hpp> //suppress_unused_warning
-
+#include <opencv2/gapi/gtransform.hpp>
 
 namespace cv {
 
@@ -57,12 +56,15 @@ namespace detail
     //
     namespace
     {
-
         template<typename T> struct Yield;
         template<> struct Yield<cv::GMat>
         {
             static inline cv::GMat yield(cv::GCall &call, int i) { return call.yield(i); }
         };
+        template<> struct Yield<cv::GMatP>
+        {
+            static inline cv::GMatP yield(cv::GCall &call, int i) { return call.yieldP(i); }
+        };
         template<> struct Yield<cv::GScalar>
         {
             static inline cv::GScalar yield(cv::GCall &call, int i) { return call.yieldScalar(i); }
@@ -82,6 +84,7 @@ namespace detail
     //    This mapping is used to transform types to call outMeta() callback.
     template<typename T> struct MetaType;
     template<> struct MetaType<cv::GMat>    { using type = GMatDesc; };
+    template<> struct MetaType<cv::GMatP>   { using type = GMatDesc; };
     template<> struct MetaType<cv::GScalar> { using type = GScalarDesc; };
     template<typename U> struct MetaType<cv::GArray<U> > { using type = GArrayDesc; };
     template<typename T> struct MetaType    { using type = T; }; // opaque args passed as-is
@@ -168,12 +171,12 @@ namespace detail
 // GKernelType and GKernelTypeM are base classes which implement typed ::on()
 // method based on kernel signature. GKernelTypeM stands for multiple-return-value kernels
 //
-// G_TYPED_KERNEL and G_TYPED_KERNEK_M macros inherit user classes from GKernelType and
+// G_TYPED_KERNEL and G_TYPED_KERNEL_M macros inherit user classes from GKernelType and
 // GKernelTypeM respectively.
 
 template<typename K, typename... R, typename... Args>
 class GKernelTypeM<K, std::function<std::tuple<R...>(Args...)> >:
-        public detail::MetaHelper<K, std::tuple<Args...>, std::tuple<R...> >
+        public detail::MetaHelper<K, std::tuple<Args...>, std::tuple<R...>>
 {
     template<int... IIs>
     static std::tuple<R...> yield(cv::GCall &call, detail::Seq<IIs...>)
@@ -197,7 +200,7 @@ template<typename, typename> class GKernelType;
 
 template<typename K, typename R, typename... Args>
 class GKernelType<K, std::function<R(Args...)> >:
-        public detail::MetaHelper<K, std::tuple<Args...>, R >
+        public detail::MetaHelper<K, std::tuple<Args...>, R>
 {
 public:
     using InArgs  = std::tuple<Args...>;
@@ -238,18 +241,11 @@ public:
 #define G_TYPED_KERNEL_M(Class, API, Id)                                    \
     G_ID_HELPER_BODY(Class, Id)                                             \
     struct Class final: public cv::GKernelTypeM<Class, std::function API >, \
-                        public detail::G_ID_HELPER_CLASS(Class)             \
+                        public detail::G_ID_HELPER_CLASS(Class)
 // {body} is to be defined by user
 
 namespace cv
 {
-// Declare <unite> in cv:: namespace
-enum class unite_policy
-{
-    REPLACE,
-    KEEP
-};
-
 namespace gapi
 {
     // Prework: model "Device" API before it gets to G-API headers.
@@ -298,42 +294,16 @@ namespace gapi {
      * @{
      */
 
-    // Lookup order is in fact a vector of Backends to traverse during look-up
-    /**
-     * @brief Priority list of backends to use during kernel
-     *   resolution process.
-     *
-     * Priority is descending -- the first backend in the list has the
-     * top priority, and the last one has the lowest priority.
-     *
-     * If there's multiple implementations available for a kernel at
-     * the moment of graph compilation, a kernel (and thus a backend)
-     * will be selected according to this order (if the parameter is passed).
-     *
-     * Default order is not specified (and by default, only
-     * CPU(OpenCV) backend is involved in graph compilation).
-     */
-    using GLookupOrder = std::vector<GBackend>;
-    /**
-     * @brief Create a backend lookup order -- priority list of
-     * backends to use during graph compilation process.
-     *
-     * @sa GLookupOrder, @ref gapi_std_backends
-     */
-    inline GLookupOrder lookup_order(std::initializer_list<GBackend> &&list)
-    {
-        return GLookupOrder(std::move(list));
-    }
-
     // FIXME: Hide implementation
     /**
      * @brief A container class for heterogeneous kernel
-     * implementation collections.
+     * implementation collections and graph transformations.
      *
      * GKernelPackage is a special container class which stores kernel
-     * _implementations_. Objects of this class are created and passed
-     * to cv::GComputation::compile() to specify which kernels to use
-     * in the compiled graph. GKernelPackage may contain kernels of
+     * _implementations_ and graph _transformations_. Objects of this class
+     * are created and passed to cv::GComputation::compile() to specify
+     * which kernels to use and which transformations to apply in the
+     * compiled graph. GKernelPackage may contain kernels of
      * different backends, e.g. be heterogeneous.
      *
      * The most easy way to create a kernel package is to use function
@@ -345,23 +315,23 @@ namespace gapi {
      * with an empty package (created with the default constructor)
      * and then by populating it with kernels via call to
      * GKernelPackage::include(). Note this method is also a template
-     * one since G-API kernel implementations are _types_, not objects.
+     * one since G-API kernel and transformation implementations are _types_,
+     * not objects.
      *
      * Finally, two kernel packages can be combined into a new one
-     * with function cv::gapi::combine(). There are different rules
-     * apply to this process, see also cv::gapi::unite_policy for
-     * details.
+     * with function cv::gapi::combine().
      */
     class GAPI_EXPORTS GKernelPackage
     {
-        /// @private
-        using S = std::unordered_map<std::string, GKernelImpl>;
 
         /// @private
-        using M = std::unordered_map<GBackend, S>;
+        using M = std::unordered_map<std::string, std::pair<GBackend, GKernelImpl>>;
 
         /// @private
-        M m_backend_kernels;
+        M m_id_kernels;
+
+        /// @private
+        std::vector<GTransform> m_transformations;
 
     protected:
         /// @private
@@ -373,30 +343,64 @@ namespace gapi {
         // Remove ALL implementations of the given API (identified by ID)
         void removeAPI(const std::string &id);
 
+        /// @private
+        // Partial include() specialization for kernels
+        template <typename KImpl>
+        typename std::enable_if<(std::is_base_of<detail::KernelTag, KImpl>::value), void>::type
+        includeHelper()
+        {
+            auto backend     = KImpl::backend();
+            auto kernel_id   = KImpl::API::id();
+            auto kernel_impl = GKernelImpl{KImpl::kernel()};
+            removeAPI(kernel_id);
+
+            m_id_kernels[kernel_id] = std::make_pair(backend, kernel_impl);
+        }
+
+        /// @private
+        // Partial include() specialization for transformations
+        template <typename TImpl>
+        typename std::enable_if<(std::is_base_of<detail::TransformTag, TImpl>::value), void>::type
+        includeHelper()
+        {
+            m_transformations.emplace_back(TImpl::transformation());
+        }
+
     public:
         /**
-         * @brief Returns total number of kernels in the package
-         * (accross all backends included)
+         * @brief Returns total number of kernels
+         * in the package (across all backends included)
          *
          * @return a number of kernels in the package
          */
         std::size_t size() const;
 
+        /**
+         * @brief Returns vector of transformations included in the package
+         *
+         * @return vector of transformations included in the package
+         */
+        const std::vector<GTransform>& get_transformations() const;
+
         /**
          * @brief Test if a particular kernel _implementation_ KImpl is
          * included in this kernel package.
          *
          * @sa includesAPI()
          *
+         * @note cannot be applied to transformations
+         *
          * @return true if there is such kernel, false otherwise.
          */
         template<typename KImpl>
         bool includes() const
         {
-            const auto set_iter = m_backend_kernels.find(KImpl::backend());
-            return (set_iter != m_backend_kernels.end())
-                ? (set_iter->second.count(KImpl::API::id()) > 0)
-                : false;
+            static_assert(std::is_base_of<detail::KernelTag, KImpl>::value,
+                          "includes() can be applied to kernels only");
+
+            auto kernel_it = m_id_kernels.find(KImpl::API::id());
+            return kernel_it != m_id_kernels.end() &&
+                   kernel_it->second.first == KImpl::backend();
         }
 
         /**
@@ -434,47 +438,33 @@ namespace gapi {
         }
 
         /**
-         * @brief Find a kernel (by its API), given the look-up order.
+         * @brief Find a kernel (by its API)
          *
-         * If order is empty, returns first suitable implementation.
+         * Returns implementation corresponding id.
          * Throws if nothing found.
          *
          * @return Backend which hosts matching kernel implementation.
          *
-         * @sa cv::gapi::lookup_order
          */
         template<typename KAPI>
-        GBackend lookup(const GLookupOrder &order = {}) const
+        GBackend lookup() const
         {
-            return lookup(KAPI::id(), order).first;
+            return lookup(KAPI::id()).first;
         }
 
         /// @private
         std::pair<cv::gapi::GBackend, cv::GKernelImpl>
-        lookup(const std::string &id, const GLookupOrder &order = {}) const;
+        lookup(const std::string &id) const;
 
         // FIXME: No overwrites allowed?
         /**
-         * @brief Put a new kernel implementation KImpl into package.
-         *
-         * @param up unite policy to use. If the package has already
-         * implementation for this kernel (probably from another
-         * backend), and cv::unite_policy::KEEP is passed, the
-         * existing implementation remains in package; on
-         * cv::unite_policy::REPLACE all other existing
-         * implementations are first dropped from the package.
+         * @brief Put a new kernel implementation or a new transformation
+         * KImpl into the package.
          */
         template<typename KImpl>
-        void include(const cv::unite_policy up = cv::unite_policy::KEEP)
+        void include()
         {
-            auto backend     = KImpl::backend();
-            auto kernel_id   = KImpl::API::id();
-            auto kernel_impl = GKernelImpl{KImpl::kernel()};
-            if (up == cv::unite_policy::REPLACE) removeAPI(kernel_id);
-            else GAPI_Assert(up == cv::unite_policy::KEEP);
-
-            // Regardless of the policy, store new impl in its storage slot.
-            m_backend_kernels[backend][kernel_id] = std::move(kernel_impl);
+            includeHelper<KImpl>();
         }
 
         /**
@@ -487,36 +477,27 @@ namespace gapi {
         // TODO: Doxygen bug -- it wants me to place this comment
         // here, not below.
         /**
-         * @brief Create a new package based on `lhs` and `rhs`,
-         * with unity policy defined by `policy`.
+         * @brief Create a new package based on `lhs` and `rhs`.
          *
          * @param lhs "Left-hand-side" package in the process
          * @param rhs "Right-hand-side" package in the process
-         * @param policy Unite policy which is used in case of conflicts
-         * -- when the same kernel API is implemented in both packages by
-         * different backends; cv::unite_policy::KEEP keeps both
-         * implementation in the resulting package, while
-         * cv::unite_policy::REPLACE gives precedence two kernels from
-         * "Right-hand-side".
-         *
          * @return a new kernel package.
          */
         friend GAPI_EXPORTS GKernelPackage combine(const GKernelPackage  &lhs,
-                                                   const GKernelPackage  &rhs,
-                                                   const cv::unite_policy policy);
+                                                   const GKernelPackage  &rhs);
     };
 
     /**
      * @brief Create a kernel package object containing kernels
-     * specified in variadic template argument.
+     * and transformations specified in variadic template argument.
      *
-     * In G-API, kernel implementations are _types_. Every backend has
-     * its own kernel API (like GAPI_OCV_KERNEL() and
+     * In G-API, kernel implementations and transformations are _types_.
+     * Every backend has its own kernel API (like GAPI_OCV_KERNEL() and
      * GAPI_FLUID_KERNEL()) but all of that APIs define a new type for
      * each kernel implementation.
      *
      * Use this function to pass kernel implementations (defined in
-     * either way) to the system. Example:
+     * either way) and transformations to the system. Example:
      *
      * @snippet modules/gapi/samples/api_ref_snippets.cpp kernels_snippet
      *
@@ -526,6 +507,10 @@ namespace gapi {
      */
     template<typename... KK> GKernelPackage kernels()
     {
+        // FIXME: currently there is no check that transformations' signatures are unique
+        // and won't be any intersection in graph compilation stage
+        static_assert(detail::all_unique<typename KK::API...>::value, "Kernels API must be unique");
+
         GKernelPackage pkg;
 
         // For those who wonder - below is a trick to call a number of
@@ -534,7 +519,6 @@ namespace gapi {
         // Just note that `f(),a` always equals to `a` (with f() called!)
         // and parentheses are used to hide function call in the expanded sequence.
         // Leading 0 helps to handle case when KK is an empty list (kernels<>()).
-
         int unused[] = { 0, (pkg.include<KK>(), 0)... };
         cv::util::suppress_unused_warning(unused);
         return pkg;
@@ -543,8 +527,17 @@ namespace gapi {
     /** @} */
 
     GAPI_EXPORTS GKernelPackage combine(const GKernelPackage  &lhs,
-                                        const GKernelPackage  &rhs,
-                                        const cv::unite_policy policy);
+                                        const GKernelPackage  &rhs);
+    /**
+     * @brief cv::use_only() is a special combinator which hints G-API to use only
+     * kernels specified in cv::GComputation::compile() (and not to extend kernels available by
+     * default with that package).
+     */
+    struct GAPI_EXPORTS use_only
+    {
+        GKernelPackage pkg;
+    };
+
 } // namespace gapi
 
 namespace detail
@@ -553,9 +546,10 @@ namespace detail
     {
         static const char* tag() { return "gapi.kernel_package"; }
     };
-    template<> struct CompileArgTag<cv::gapi::GLookupOrder>
+
+    template<> struct CompileArgTag<cv::gapi::use_only>
     {
-        static const char* tag() { return "gapi.lookup_order"; }
+        static const char* tag() { return "gapi.use_only"; }
     };
 } // namespace detail
 } // namespace cv
